import { execute, graphql } from 'graphql';
import request from '../entrypoints/main.js';
import { Schema } from '../Schema/__.js';
import { readMaybeThunk } from '../Schema/core/helpers.js';
import * as CustomScalars from './customScalars.js';
import { toDocumentExpression } from './document.js';
import { SelectionSet } from './SelectionSet/__.js';
export const create = (input) => {
    const parentInput = input;
    const returnMode = input.returnMode ?? `data`;
    const runHookable = (name, input, fn) => {
        return parentInput.hooks?.[name](input, fn) ?? fn(input);
    };
    const executeDocumentExpression = async ({ document, variables, operationName }) => {
        let result;
        if (input.schema instanceof URL || typeof input.schema === `string`) {
            // todo return errors too
            const data = await request({
                url: new URL(input.schema).href, // todo allow relative urls - what does fetch in node do?
                requestHeaders: input.headers,
                document,
                variables,
                // todo use operationName
            });
            result = { data };
        }
        else {
            if (typeof document === `string`) {
                result = await graphql({
                    schema: input.schema,
                    source: document,
                    // contextValue: createContextValue(), // todo
                    variableValues: variables,
                    operationName,
                });
            }
            else if (typeof document === `object`) {
                result = await execute({
                    schema: input.schema,
                    document,
                    // contextValue: createContextValue(), // todo
                    variableValues: variables,
                    operationName,
                });
            }
            else {
                throw new Error(`Unsupported GraphQL document type: ${String(document)}`);
            }
        }
        return result;
    };
    const executeRootObject = (rootTypeName) => async (documentObject) => {
        const rootIndex = input.schemaIndex.Root[rootTypeName];
        if (!rootIndex)
            throw new Error(`Root type not found: ${rootTypeName}`);
        const documentObjectEncoded = runHookable(`documentEncode`, { rootIndex, documentObject }, ({ rootIndex, documentObject }) => CustomScalars.encode({ index: rootIndex, documentObject }));
        const documentString = SelectionSet.selectionSet(documentObjectEncoded);
        // todo variables
        const result = await executeDocumentExpression({ document: documentString });
        // @ts-expect-error todo make global available in TS...
        if (result.errors && (result.errors.length > 0))
            throw new AggregateError(result.errors); // eslint-disable-line
        // todo check for errors
        const dataDecoded = CustomScalars.decode(rootIndex, result.data);
        return { ...result, data: dataDecoded };
    };
    const rootTypeNameToOperationName = {
        Query: `query`,
        Mutation: `mutation`,
        Subscription: `subscription`,
    };
    const executeRootTypeFieldSelection = (rootTypeName, key) => {
        return async (argsOrSelectionSet) => {
            const type = readMaybeThunk(
            // eslint-disable-next-line
            // @ts-ignore excess depth error
            Schema.Output.unwrapToNamed(readMaybeThunk(input.schemaIndex.Root[rootTypeName]?.fields[key]?.type)));
            if (!type)
                throw new Error(`${rootTypeName} field not found: ${String(key)}`); // eslint-disable-line
            // @ts-expect-error fixme
            const isSchemaScalarOrTypeName = type.kind === `Scalar` || type.kind === `typename`; // todo fix type here, its valid
            const isSchemaHasArgs = Boolean(input.schemaIndex.Root[rootTypeName]?.fields[key]?.args);
            const documentObject = {
                [rootTypeNameToOperationName[rootTypeName]]: {
                    [key]: isSchemaScalarOrTypeName
                        ? isSchemaHasArgs && argsOrSelectionSet ? { $: argsOrSelectionSet } : true
                        : argsOrSelectionSet,
                },
            };
            const result = await rootObjectExecutors[rootTypeName](documentObject);
            const resultHandled = handleReturn(result);
            // @ts-expect-error make this type safe?
            return returnMode === `data` ? resultHandled[key] : resultHandled;
        };
    };
    const handleReturn = (result) => {
        switch (returnMode) {
            case `data`: {
                if (result.errors && result.errors.length > 0) {
                    // @ts-expect-error fixme -- get this global available
                    throw new AggregateError(result.errors); // eslint-disable-line
                }
                return result.data;
            }
            default: {
                return result;
            }
        }
    };
    const rootObjectExecutors = {
        Mutation: executeRootObject(`Mutation`),
        Query: executeRootObject(`Query`),
        Subscription: executeRootObject(`Subscription`),
    };
    const createRootTypeMethods = (rootTypeName) => new Proxy({}, {
        get: (_, key) => {
            if (typeof key === `symbol`)
                throw new Error(`Symbols not supported.`);
            if (key === `$batch`) {
                return async (selectionSetOrIndicator) => {
                    const result = await rootObjectExecutors[rootTypeName]({
                        [rootTypeNameToOperationName[rootTypeName]]: selectionSetOrIndicator,
                    });
                    return handleReturn(result);
                };
            }
            else {
                return executeRootTypeFieldSelection(rootTypeName, key);
            }
        },
    });
    // @ts-expect-error ignoreme
    const client = {
        raw: async (document, variables, operationName) => {
            return await executeDocumentExpression({ document, variables, operationName });
        },
        document: (documentObject) => {
            return {
                run: async (operationName) => {
                    // todo this does not support custom scalars
                    const documentExpression = toDocumentExpression(documentObject);
                    const result = await executeDocumentExpression({
                        document: documentExpression,
                        operationName,
                        // todo variables
                    });
                    return handleReturn(result);
                },
            };
        },
        query: createRootTypeMethods(`Query`),
        mutation: createRootTypeMethods(`Mutation`),
        // todo
        // subscription: async () => {},
    };
    return client;
};
//# sourceMappingURL=client.js.map